@@ -24,6 +24,7 @@ from typing import (
 from urllib.parse import urlencode, urlparse
 
 import requests
+import semver
 from gi.repository import Gtk
 
 from .api_objects import Directory, Response
@@ -145,6 +146,12 @@ class SubsonicAdapter(Adapter):
                         timeout=2,
                         is_exponential_backoff_ping=True,
                     )
+                except requests.exceptions.SSLError:
+                    errors["__ping__"] = (
+                        "<b>Error connecting to the server.</b>\n"
+                        "An SSL error occurred while connecting to the server.\n"
+                        "You may need to explicitly specify http://."
+                    )
                 except requests.ConnectionError:
                     errors["__ping__"] = (
                         "<b>Unable to connect to the server.</b>\n"
@@ -152,7 +159,7 @@ class SubsonicAdapter(Adapter):
                     )
                 except ServerError as e:
                     errors["__ping__"] = (
-                        "<b>Error connecting in to the server.</b>\n"
+                        "<b>Error connecting to the server.</b>\n"
                         f"Error {e.status_code}: {str(e)}"
                     )
                 except Exception as e:
@@ -201,16 +208,16 @@ class SubsonicAdapter(Adapter):
             self.hostname = "https://" + self.hostname
 
         self.username = config["username"]
-        self.password = config.get_secret("password")
+        self.password = cast(str, config.get_secret("password"))
         self.verify_cert = config["verify_cert"]
 
         self.is_shutting_down = False
+        self._ping_process: Optional[multiprocessing.Process] = None
+        self._version = multiprocessing.Array("c", 20)
+        self._offline_mode = False
 
         # TODO (#112): support XML?
 
-    _ping_process: Optional[multiprocessing.Process] = None
-    _offline_mode = False
-
     def initial_sync(self):
         # Try to ping the server five times using exponential backoff (2^5 = 32s).
         self._exponential_backoff(5)
@@ -262,27 +269,41 @@ class SubsonicAdapter(Adapter):
     def ping_status(self) -> bool:
         return self._server_available.value
 
-    # TODO (#199) make these way smarter
-    can_get_playlists = True
-    can_get_playlist_details = True
     can_create_playlist = True
-    can_update_playlist = True
     can_delete_playlist = True
-    can_get_cover_art_uri = True
-    can_get_song_uri = True
-    can_stream = True
-    can_get_song_details = True
-    can_scrobble_song = True
-    can_get_artists = True
-    can_get_artist = True
-    can_get_ignored_articles = True
-    can_get_albums = True
     can_get_album = True
+    can_get_albums = True
+    can_get_artist = True
+    can_get_artists = True
+    can_get_cover_art_uri = True
     can_get_directory = True
-    can_get_genres = True
-    can_get_play_queue = True
-    can_save_play_queue = True
+    can_get_ignored_articles = True
+    can_get_playlist_details = True
+    can_get_playlists = True
+    can_get_song_details = True
+    can_get_song_file_uri = True
+    can_get_song_stream_uri = True
+    can_scrobble_song = True
     can_search = True
+    can_stream = True
+    can_update_playlist = True
+
+    def version_at_least(self, version: str) -> bool:
+        if not self._version.value:
+            return False
+        return semver.VersionInfo.parse(self._version.value.decode()) >= version
+
+    @property
+    def can_get_genres(self) -> bool:
+        return self.version_at_least("1.9.0")
+
+    @property
+    def can_get_play_queue(self) -> bool:
+        return self.version_at_least("1.12.0")
+
+    @property
+    def can_save_play_queue(self) -> bool:
+        return self.version_at_least("1.12.0")
 
     _schemes = None
 
@@ -292,18 +313,22 @@ class SubsonicAdapter(Adapter):
             self._schemes = (urlparse(self.hostname)[0],)
         return self._schemes
 
-    # TODO (#203) make this way smarter
-    supported_artist_query_types = {
-        AlbumSearchQuery.Type.RANDOM,
-        AlbumSearchQuery.Type.NEWEST,
-        AlbumSearchQuery.Type.FREQUENT,
-        AlbumSearchQuery.Type.RECENT,
-        AlbumSearchQuery.Type.STARRED,
-        AlbumSearchQuery.Type.ALPHABETICAL_BY_NAME,
-        AlbumSearchQuery.Type.ALPHABETICAL_BY_ARTIST,
-        AlbumSearchQuery.Type.YEAR_RANGE,
-        AlbumSearchQuery.Type.GENRE,
-    }
+    @property
+    def supported_artist_query_types(self) -> Set[AlbumSearchQuery.Type]:
+        supported = {
+            AlbumSearchQuery.Type.RANDOM,
+            AlbumSearchQuery.Type.NEWEST,
+            AlbumSearchQuery.Type.FREQUENT,
+            AlbumSearchQuery.Type.RECENT,
+            AlbumSearchQuery.Type.STARRED,
+            AlbumSearchQuery.Type.ALPHABETICAL_BY_NAME,
+            AlbumSearchQuery.Type.ALPHABETICAL_BY_ARTIST,
+        }
+        if self.version_at_least("1.10.1"):
+            supported.add(AlbumSearchQuery.Type.YEAR_RANGE)
+            supported.add(AlbumSearchQuery.Type.GENRE)
+
+        return supported
 
     # Helper mothods for making requests
     # ==================================================================================
@@ -317,7 +342,7 @@ class SubsonicAdapter(Adapter):
             "p": self.password,
             "c": "Sublime Music",
             "f": "json",
-            "v": "1.15.0",
+            "v": self._version.value.decode() or "1.8.0",
         }
 
     def _make_url(self, endpoint: str) -> str:
@@ -412,12 +437,14 @@ class SubsonicAdapter(Adapter):
         if not subsonic_response:
             raise ServerError(500, f"{url} returned invalid JSON.")
 
-        if subsonic_response["status"] == "failed":
+        if subsonic_response["status"] != "ok":
             raise ServerError(
                 subsonic_response["error"].get("code"),
                 subsonic_response["error"].get("message"),
             )
 
+        self._version.value = subsonic_response["version"].encode()
+
         logging.debug(f"Response from {url}: {subsonic_response}")
         return Response.from_dict(subsonic_response)
 
@@ -511,10 +538,14 @@ class SubsonicAdapter(Adapter):
         params = {"id": cover_art_id, "size": size, **self._get_params()}
         return self._make_url("getCoverArt") + "?" + urlencode(params)
 
-    def get_song_uri(self, song_id: str, scheme: str, stream: bool = False) -> str:
+    def get_song_file_uri(self, song_id: str, schemes: Iterable[str]) -> str:
+        assert any(s in schemes for s in self.supported_schemes)
         params = {"id": song_id, **self._get_params()}
-        endpoint = "stream" if stream else "download"
-        return self._make_url(endpoint) + "?" + urlencode(params)
+        return self._make_url("download") + "?" + urlencode(params)
+
+    def get_song_stream_uri(self, song_id: str) -> str:
+        params = {"id": song_id, **self._get_params()}
+        return self._make_url("stream") + "?" + urlencode(params)
 
     def get_song_details(self, song_id: str) -> API.Song:
         song = self._get_json(self._make_url("getSong"), id=song_id).song
@@ -538,9 +569,14 @@ class SubsonicAdapter(Adapter):
     def get_artist(self, artist_id: str) -> API.Artist:
         artist = self._get_json(self._make_url("getArtist"), id=artist_id).artist
         assert artist, f"Error getting artist {artist_id}"
-        artist.augment_with_artist_info(
-            self._get_json(self._make_url("getArtistInfo2"), id=artist_id).artist_info
-        )
+        if self.version_at_least("1.11.0"):
+            try:
+                artist_info = self._get_json(
+                    self._make_url("getArtistInfo2"), id=artist_id
+                )
+                artist.augment_with_artist_info(artist_info.artist_info)
+            except Exception:
+                pass
         return artist
 
     def get_ignored_articles(self) -> Set[str]:
@@ -651,6 +687,7 @@ class SubsonicAdapter(Adapter):
         self._get(
             self._make_url("savePlayQueue"),
             id=song_ids,
+            timeout=2,
             current=song_ids[current_song_index]
             if current_song_index is not None
             else None,
